# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Definitions for mapping classes.
"""

__all__ = [
    "FrozenNamespace",
    "Namespace",
    "OpenFrozenNamespace",
    "OpenNamespace",
]


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue

from .slots import Slots


class Namespace(Slots):
    """
    A simple wrapper for turning mappings into Namespace objects that
    allow dotted lookup and assignment on those mappings.

    Masked repr.
    Mutable instance.

     _____________________________________
    |                                     |
    |   Stability of Assignment Styles:   |
    |_____________________________________|


    from aiootp.commons.namespaces import Namespace

    class HybridNamespace(Namespace):
        __slots__ = ("attr",)

    hybrid = HybridNamespace()

    ✔ hybrid.attr = "value"                # supported
    ✔ hybrid["attr"] = "value"             # supported
    ✔ setattr(hybrid, "attr", "value")     # supported

    ❌ hybrid.__dict__["attr"] = "value"    # unsupported


    # See: https://github.com/rmlibre/aiootp/pull/11
    """

    __slots__ = ("__dict__",)


class OpenNamespace(Namespace):
    """
    A version of the `Namespace` class which doesn't omit instance
    repr's by default.

    Unmasked repr.
    Mutable instance.
    """

    __slots__ = ()

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


class FrozenNamespace(Namespace):
    """
    A simple wrapper for turning mappings into Namespace objects that
    allow dotted lookup and assignment on those mappings. Also, provides
    utilities for copying mappings into other containers, & turning
    mappings into stand-alone, first-class modules.

    Masked repr.
    Immutable instance. (set once)
    """

    __slots__ = ()

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Denies setting attributes after they have already been set.
        """
        if name in self:
            raise Issue.cant_reassign_attribute(name)

        super().__setitem__(name, value)

    def __delitem__(self, name: str, /) -> None:
        """
        Denies deleting attributes.
        """
        raise Issue.cant_deassign_attribute(name)


class OpenFrozenNamespace(FrozenNamespace):
    """
    A version of the `FrozenNamespace` class which doesn't omit instance
    repr's by default.

    Unmasked repr.
    Immutable instance. (set once)
    """

    __slots__ = ()

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


module_api = dict(
    FrozenNamespace=t.add_type(FrozenNamespace),
    Namespace=t.add_type(Namespace),
    OpenFrozenNamespace=t.add_type(OpenFrozenNamespace),
    OpenNamespace=t.add_type(OpenNamespace),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A setuptools build script run automatically by pyproject.toml with:
``python3 -m build``
"""

from setuptools import setup


if __name__ == "__main__":
    setup()



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Manages the early secrets used by the package.
"""

__all__ = []


from collections import deque

from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp._paths import SecurePath, read_salt_file
from aiootp.asynchs import asleep, ns_counter
from aiootp.generics import Domains

from .simple import token_bits, token_bytes


# initialize rudimentary global entropy pool
_int_pool = deque((token_bits(256), token_bits(256)), maxlen=2)


# create and/or retrieve the device seed stored on the filesystem
_package_seed_name = Domains.encode_constant(b"package_seed", size=16)
_package_seed_path = SecurePath(key=_package_seed_name, _admin=True)
_package_seed = read_salt_file(_package_seed_path)


# prepare global salted multiply values
_MOD = 1 << 256
_offset = (0x99 << 256) | token_bits(256)  # ensure always a non-zero offset
_mix = int.from_bytes(_package_seed, BIG)
_seed = int.from_bytes(token_bytes(32), BIG)


async def _asalt_multiply(*numbers: int, pool: t.Sequence[bytes]) -> int:
    """
    Allows for non-commutative multiplication. This assists pseudo-
    random number generators in turning combinations of low entropy
    number sources into permutations. This greatly increases the amount
    of knowledge an attacker must have to perform pre-image or known-
    plaintext attacks on the calculations usings the hashes of those
    permutations.
    """
    global _mix, _seed, _offset

    await asleep()
    start = ns_counter() * _offset
    mix = _mix ^ abs(sum((start, *_int_pool, *numbers)))
    mix = _mix = (mix ^ int.from_bytes(pool[0], BIG)) % _MOD
    _seed ^= mix
    _offset ^= _seed
    for number in numbers:
        await asleep()
        mix += _offset
        start = (start ^ mix) * (number ^ mix)
    return _seed ^ ((start ^ mix) % _MOD)


def _salt_multiply(*numbers: int, pool: t.Sequence[bytes]) -> int:
    """
    Allows for non-commutative multiplication. This assists pseudo-
    random number generators in turning combinations of low entropy
    number sources into permutations. This greatly increases the amount
    of knowledge an attacker must have to perform pre-image or known-
    plaintext attacks on the calculations usings the hashes of those
    permutations.
    """
    global _mix, _seed, _offset

    start = ns_counter() * _offset
    mix = _mix ^ abs(sum((start, *_int_pool, *numbers)))
    mix = _mix = (mix ^ int.from_bytes(pool[0], BIG)) % _MOD
    _seed ^= mix
    _offset ^= _seed
    for number in numbers:
        mix += _offset
        start = (start ^ mix) * (number ^ mix)
    return _seed ^ ((start ^ mix) % _MOD)


async def _asalt(
    *, pool: t.Sequence[bytes], gadget: t.EntropyHashingType
) -> int:
    """
    Returns a 256-bit number derived from cached & ratcheted system
    entropy.
    """
    entropy = gadget.hash(size=32)
    _int_pool.appendleft(int.from_bytes(entropy, BIG))
    return await _asalt_multiply(*_int_pool, pool=pool)


def _salt(*, pool: t.Sequence[bytes], gadget: t.EntropyHashingType) -> int:
    """
    Returns a 256-bit number derived from cached & ratcheted system
    entropy.
    """
    entropy = gadget.hash(size=32)
    _int_pool.appendleft(int.from_bytes(entropy, BIG))
    return _salt_multiply(*_int_pool, pool=pool)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Canonicalization utilities.
"""

__all__ = [
    "abytes_are_equal",
    "acanonical_pack",
    "acanonical_unpack",
    "adecode_items",
    "aencode_items",
    "aencode_key",
    "afullblock_ljust",
    "bytes_are_equal",
    "canonical_pack",
    "canonical_unpack",
    "decode_items",
    "fullblock_ljust",
    "encode_items",
    "encode_key",
    "test_canonical_padding",
]


import io
from math import ceil
from collections import deque
from hmac import compare_digest as bytes_are_equal

from aiootp._typing import Typing as t
from aiootp._constants import LITTLE, BIG, INT_BYTES
from aiootp._exceptions import Issue, CanonicalIssue
from aiootp._gentools import arange
from aiootp.asynchs import asleep


async def abytes_are_equal(value_0: bytes, value_1: bytes) -> bool:
    """
    Tests if two bytes values are equal with a simple & fast timing-safe
    comparison function from the `hmac` module.
    """
    await asleep()
    return bytes_are_equal(value_0, value_1)


async def afullblock_ljust(
    data: bytes, blocksize: int, *, pad: bytes = b"\x00"
) -> bytes:
    """
    Right pads a `bytes`-type `data` value to a multiple of the
    integer `blocksize` with `pad` characters.
    """
    await asleep()
    return data.ljust(blocksize * ceil(len(data) / blocksize), pad)


def fullblock_ljust(
    data: bytes, blocksize: int, *, pad: bytes = b"\x00"
) -> bytes:
    """
    Right pads a `bytes`-type `data` value to a multiple of the
    integer `blocksize` with `pad` characters.
    """
    return data.ljust(blocksize * ceil(len(data) / blocksize), pad)


async def aencode_key(
    key: bytes, blocksize: int, *, pad: bytes = b"\x00"
) -> bytes:
    """
    Returns a symmetric `key` canonically encoded so as to be
    prepended to other canonically encoded data for use in hash
    functions.
    """
    if not key:
        raise Issue.value_must("key", "be supplied for encoding")
    key_length = len(key).to_bytes(8, LITTLE)
    bsize = blocksize.to_bytes(8, LITTLE)
    return await afullblock_ljust(
        b"".join((key, key_length, bsize, pad)), blocksize, pad=pad
    )


def encode_key(
    key: bytes, blocksize: int, *, pad: bytes = b"\x00"
) -> bytes:
    """
    Returns a symmetric `key` canonically encoded so as to be
    prepended to other canonically encoded data for use in hash
    functions.
    """
    if not key:
        raise Issue.value_must("key", "be supplied for encoding")
    key_length = len(key).to_bytes(8, LITTLE)
    bsize = blocksize.to_bytes(8, LITTLE)
    return fullblock_ljust(
        b"".join((key, key_length, bsize, pad)), blocksize, pad=pad
    )


async def aencode_items(
    *items: bytes, int_bytes: int = INT_BYTES
) -> t.AsyncGenerator[bytes, None]:
    """
    Yields each item in `items` with encoded length metadata attached.
    """
    yield int_bytes.to_bytes(1, BIG)
    yield len(items).to_bytes(int_bytes, BIG)
    for item in items:
        await asleep()
        yield len(item).to_bytes(int_bytes, BIG) + item


def encode_items(
    *items: bytes, int_bytes: int = INT_BYTES
) -> t.Generator[bytes, None, None]:
    """
    Yields each item in `items` with encoded length metadata attached.
    """
    yield int_bytes.to_bytes(1, BIG)
    yield len(items).to_bytes(int_bytes, BIG)
    for item in items:
        yield len(item).to_bytes(int_bytes, BIG) + item


async def acanonical_pack(
    *items: bytes,
    blocksize: int = 1,
    pad: bytes = b"\x00",
    int_bytes: int = INT_BYTES,
) -> bytes:
    """
    Returns a joined iterable of bytes-type `items` with encoded length
    metadata of the iterable & each item attached. The result is right-
    padded with `pad` to a multiple of the `blocksize`. This can be used
    to prevent canonicalization & length extension attacks when hashing
    arbitrary collections of inputs.

    https://soatok.blog/2021/07/30/canonicalization-attacks-against-macs
    -and-signatures/

     _____________________________________
    |                                     |
    |           Format Diagram:           |
    |_____________________________________|

    I = `int_bytes`
    |----  len(result) % blocksize == 0, right padded with `pad`  -----|
     __________________________________________________________________
    |                   |               |                            | |
    |   len(Iterable)   |   len(Item)   |            Item            | |
    |                   |               |                            | |
    |     I-bytes       |    I-bytes    |           X-bytes          | |
    |     = W + 2       |    = X        |                            | |
    |___________________|_______________|____________________________| |
    |                   |                                            | |
    |   1 x at start    |           W x once for each item           | |
    |___________________|____________________________________________|_|
    """
    blocksize_blob = blocksize.to_bytes(int_bytes, BIG)
    items = [
        item
        async for item in aencode_items(
            blocksize_blob, pad, *items, int_bytes=int_bytes
        )
    ]
    return fullblock_ljust(b"".join(items), blocksize, pad=pad)


def canonical_pack(
    *items: bytes,
    blocksize: int = 1,
    pad: bytes = b"\x00",
    int_bytes: int = INT_BYTES,
) -> bytes:
    """
    Returns a joined iterable of bytes-type `items` with encoded length
    metadata of the iterable & each item attached. The result is right-
    padded with `pad` to a multiple of the `blocksize`. This can be used
    to prevent canonicalization & length extension attacks when hashing
    arbitrary collections of inputs.

    https://soatok.blog/2021/07/30/canonicalization-attacks-against-macs
    -and-signatures/

     _____________________________________
    |                                     |
    |           Format Diagram:           |
    |_____________________________________|

    I = `int_bytes`
    |----  len(result) % blocksize == 0, right padded with `pad`  -----|
     __________________________________________________________________
    |                   |               |                            | |
    |   len(Iterable)   |   len(Item)   |            Item            | |
    |                   |               |                            | |
    |     I-bytes       |    I-bytes    |           X-bytes          | |
    |     = W + 2       |    = X        |                            | |
    |___________________|_______________|____________________________| |
    |                   |                                            | |
    |   1 x at start    |           W x once for each item           | |
    |___________________|____________________________________________|_|
    """
    blocksize_blob = blocksize.to_bytes(int_bytes, BIG)
    items = encode_items(blocksize_blob, pad, *items, int_bytes=int_bytes)
    return fullblock_ljust(b"".join(items), blocksize, pad=pad)


async def adecode_items(
    read: t.Callable[[int], bytes], item_count: int, int_bytes: int
) -> t.AsyncGenerator[bytes, None]:
    """
    Extracts each size-item pair from the `read` callable, which
    outputs a number of canonically encoded bytes equal to the integer
    it receives as an argument. Yields all `item_count` number of
    items one at a time if each item's length matches its declared
    length, otherwise raises `CanonicalEncodingError`. `int_bytes` is
    the number of bytes that were used to encode item lengths.
    """
    async for _ in arange(item_count):
        item_size = int.from_bytes(read(int_bytes), BIG)
        item = read(item_size)
        if len(item) != item_size:
            raise CanonicalIssue.item_length_mismatch()
        yield item


def decode_items(
    read: t.Callable[[int], bytes], item_count: int, int_bytes: int
) -> t.Generator[bytes, None, None]:
    """
    Extracts each size-item pair from the `read` callable, which
    outputs a number of canonically encoded bytes equal to the integer
    it receives as an argument. Yields all `item_count` number of
    items one at a time if each item's length matches its declared
    length, otherwise raises `CanonicalEncodingError`. `int_bytes` is
    the number of bytes that were used to encode item lengths.
    """
    for _ in range(item_count):
        item_size = int.from_bytes(read(int_bytes), BIG)
        item = read(item_size)
        if len(item) != item_size:
            raise CanonicalIssue.item_length_mismatch()
        yield item


def test_canonical_padding(
    read: t.Callable[[int], bytes], items: t.Deque[bytes], total_size: int
) -> None:
    """
    Raises `CanonicalEncodingError` if an invalid type of padding is
    detected in the bytes-type data produced by the `read` callable.
    """
    try:
        blocksize = int.from_bytes(items.popleft(), BIG)
        pad = items.popleft()
    except IndexError as error:
        raise CanonicalIssue.missing_metadata_items() from error
    if total_size % blocksize:
        raise CanonicalIssue.data_length_blocksize_mismatch()
    elif len(pad) != 1 or not set(pad).issuperset(read()):
        raise CanonicalIssue.invalid_padding()


async def acanonical_unpack(items: bytes) -> t.Deque[bytes]:
    """
    Extracts the bytes-type values that have been canonically encoded
    into the `items` byte string. `int_bytes` is the number of bytes
    that were used to encode item lengths.
    """
    total_size = len(items)
    read = io.BytesIO(items).read
    int_bytes = int.from_bytes(read(1), BIG)
    item_count = int.from_bytes(read(int_bytes), BIG)
    if total_size < int_bytes * (item_count + 1) + 3:
        raise CanonicalIssue.inflated_size_declaration()
    items = deque(
        [item async for item in adecode_items(read, item_count, int_bytes)],
        maxlen=item_count,
    )
    test_canonical_padding(read, items, total_size)
    return items


def canonical_unpack(items: bytes) -> t.Deque[bytes]:
    """
    Extracts the bytes-type values that have been canonically encoded
    into the `items` byte string. `int_bytes` is the number of bytes
    that were used to encode item lengths.
    """
    total_size = len(items)
    read = io.BytesIO(items).read
    int_bytes = int.from_bytes(read(1), BIG)
    item_count = int.from_bytes(read(int_bytes), BIG)
    if total_size < int_bytes * (item_count + 1) + 3:
        raise CanonicalIssue.inflated_size_declaration()
    items = deque(
        decode_items(read, item_count, int_bytes), maxlen=item_count
    )
    test_canonical_padding(read, items, total_size)
    return items


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    abytes_are_equal=abytes_are_equal,
    acanonical_pack=acanonical_pack,
    acanonical_unpack=acanonical_unpack,
    adecode_items=adecode_items,
    aencode_items=aencode_items,
    aencode_key=aencode_key,
    afullblock_ljust=afullblock_ljust,
    bytes_are_equal=bytes_are_equal,
    canonical_pack=canonical_pack,
    canonical_unpack=canonical_unpack,
    decode_items=decode_items,
    fullblock_ljust=fullblock_ljust,
    encode_items=encode_items,
    encode_key=encode_key,
    test_canonical_padding=test_canonical_padding,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class TestPlaintextPadding:
    async def test_variable_length_plaintexts(self) -> None:
        for config, cipher, *_ in all_ciphers:
            Pad = Padding(config)
            assert config is Pad.config
            assert config is cipher._padding.config

            FIRST_BLOCKSIZE = (
                config.BLOCKSIZE
                - config.INNER_HEADER_BYTES
                - config.SENTINEL_BYTES
            )

            start = config.INNER_HEADER_BYTES
            for n in range(FIRST_BLOCKSIZE + config.BLOCKSIZE):
                pt = n * b"0"
                padded_pt = Pad.pad_plaintext(pt)
                sentinel = int.from_bytes(
                    padded_pt[config.SENTINEL_SLICE], BIG
                )
                end = -(sentinel if sentinel else config.BLOCKSIZE)

                # The data is prepended with a constant size inner-header &
                # appended with extra padding that is measured by the value of
                # the final byte of padded plaintext modulo the blocksize.
                assert (
                    padded_pt[start:end] == pt
                ), f"{n=} : {sentinel=} : {start=} : {end=}"


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for key derivation that facilitates domain-separation
& canonicalized input processing.
"""

__all__ = ["HashTupleKDF", "DomainKDF"]


from hashlib import shake_128

from aiootp._typing import Typing as t
from aiootp._constants import HASHER_TYPES
from aiootp._exceptions import Issue
from aiootp.commons import FrozenInstance
from aiootp.generics import Domains, ahash_bytes, hash_bytes
from aiootp.generics import acanonical_pack, canonical_pack


class HashTupleKDF(FrozenInstance):
    """
    A base type for passing KDF-related class attributes to subclasses
    in a consistent manner.
    """

    __slots__ = ("_payload",)

    _type: t.Callable[..., t.XOFType]

    def __init_subclass__(
        cls, *a: t.Any, salt_label: t.AnyStr, **kw: t.Any
    ) -> None:
        """
        Ensures subclasses can define custom key & base type hasher
        algorithms within their class bodies & have the blocksizes of
        those objects be recorded by the class correctly during
        definition time.
        """
        super().__init_subclass__(*a, **kw)
        cls._TYPE_BLOCKSIZE: int = cls._type().block_size
        cls._new_payload: t.Callable[[], t.XOFType] = cls._type(
            Domains.encode_constant(salt_label, size=cls._TYPE_BLOCKSIZE)
        ).copy

    def _initialize_payload(
        self, domain: bytes, data: t.Iterable[bytes], *, key: bytes
    ) -> None:
        """
        Canonically encodes then hashes the domain, key & first batch of
        input data to prepare a hashing object that processes all
        additional data & acts as an initial key generator subroutine
        for other key derivation procedures.
        """
        self._payload = self._new_payload()
        self._payload.update(
            hash_bytes(
                Domains.KDF,
                domain,
                *data,
                key=key + domain,
                size=self._payload.block_size,
                hasher=self._type,
            )
        )

    async def _arun_kdf(
        self,
        *data: bytes,
        size: t.Optional[int] = None,
        aad: bytes,
        obj: t.SimpleNamespace,
    ) -> bytes:
        """
        A generic procedure for key derivation using the description of
        a desired hash function, along with an XOF for KDF context
        absorbsion, state memory, & initial key production. The input
        `data`, `aad`, & domain of the KDF are canonically encoded &
        passed through the keyed-hashed under the initial key.
        """
        return await ahash_bytes(
            aad,
            *data,
            size=None if obj.digest_size else size,
            key=self._payload.digest(obj.block_size),
            hasher=obj.factory,
        )

    def _run_kdf(
        self,
        *data: bytes,
        size: t.Optional[int] = None,
        aad: bytes,
        obj: t.SimpleNamespace,
    ) -> bytes:
        """
        A generic procedure for key derivation using the description of
        a desired hash function, along with an XOF for KDF context
        absorbsion, state memory, & initial key production. The input
        `data`, `aad`, & domain of the KDF are canonically encoded &
        passed through the keyed-hashed under the initial key.
        """
        return hash_bytes(
            aad,
            *data,
            size=None if obj.digest_size else size,
            key=self._payload.digest(obj.block_size),
            hasher=obj.factory,
        )


class DomainKDF(HashTupleKDF, salt_label=b"domain_kdf_salt"):
    """
    Creates objects able to derive domain & data-specific keyed hashes.
    Payload updates are automatically canonicalized.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import DomainKDF


    kdf = DomainKDF(b"ecdhe", session.transcript, key=session.key)

    auth_key = kdf.sha3_512(aad=b"auth-key")
    encryption_key = kdf.sha3_512(aad=b"encryption-key")
    """

    __slots__ = ()

    _type: t.Callable[[bytes], t.XOFType] = shake_128

    def __init__(self, domain: bytes, *data: bytes, key: bytes) -> None:
        """
        Initializes a `domain`-specific KDF with a `key`, & associated
        payload `data`, which is canonically encoded by the class.
        """
        self._initialize_payload(domain, data, key=key)

    def copy(self) -> t.Cls:
        """
        Copies the instance state into a new object which can be updated
        separately in differing contexts.
        """
        kdf = self.__class__.__new__(self.__class__)
        kdf._payload = self._payload.copy()
        return kdf

    async def aupdate(self, *data: bytes) -> t.Self:
        """
        Canonically updates the payload object with additional values.
        Update calls, input data & the order of both, must match exactly
        to create matching KDF states in distinct instances.
        """
        if not data:
            raise Issue.value_must("update data", "not be empty")
        payload = self._payload
        payload.update(
            await acanonical_pack(*data, blocksize=payload.block_size)
        )
        return self

    def update(self, *data: bytes) -> t.Self:
        """
        Canonically updates the payload object with additional values.
        Update calls, input data & the order of both, must match exactly
        to create matching KDF states in distinct instances.
        """
        if not data:
            raise Issue.value_must("update data", "not be empty")
        payload = self._payload
        payload.update(canonical_pack(*data, blocksize=payload.block_size))
        return self

    async def asha3_256(self, *data: bytes, aad: bytes = b"") -> bytes:
        """
        Return the keyed sha3_256 hash of the instance's state.
        """
        return await self._arun_kdf(
            *data, aad=aad, obj=HASHER_TYPES["sha3_256"]
        )

    def sha3_256(self, *data: bytes, aad: bytes = b"") -> bytes:
        """
        Return the keyed sha3_256 hash of the instance's state.
        """
        return self._run_kdf(*data, aad=aad, obj=HASHER_TYPES["sha3_256"])

    async def asha3_512(self, *data: bytes, aad: bytes = b"") -> bytes:
        """
        Return the keyed sha3_512 hash of the instance's state.
        """
        return await self._arun_kdf(
            *data, aad=aad, obj=HASHER_TYPES["sha3_512"]
        )

    def sha3_512(self, *data: bytes, aad: bytes = b"") -> bytes:
        """
        Return the keyed sha3_512 hash of the instance's state.
        """
        return self._run_kdf(*data, aad=aad, obj=HASHER_TYPES["sha3_512"])

    async def ashake_128(
        self, *data: bytes, size: int, aad: bytes = b""
    ) -> bytes:
        """
        Return the keyed shake_128 hash of the instance's state.
        """
        return await self._arun_kdf(
            *data, size=size, aad=aad, obj=HASHER_TYPES["shake_128"]
        )

    def shake_128(self, *data: bytes, size: int, aad: bytes = b"") -> bytes:
        """
        Return the keyed shake_128 hash of the instance's state.
        """
        return self._run_kdf(
            *data, size=size, aad=aad, obj=HASHER_TYPES["shake_128"]
        )

    async def ashake_256(
        self, *data: bytes, size: int, aad: bytes = b""
    ) -> bytes:
        """
        Return the keyed shake_256 hash of the instance's state.
        """
        return await self._arun_kdf(
            *data, size=size, aad=aad, obj=HASHER_TYPES["shake_256"]
        )

    def shake_256(self, *data: bytes, size: int, aad: bytes = b"") -> bytes:
        """
        Return the keyed shake_256 hash of the instance's state.
        """
        return self._run_kdf(
            *data, size=size, aad=aad, obj=HASHER_TYPES["shake_256"]
        )


module_api = dict(
    DomainKDF=t.add_type(DomainKDF),
    HashTupleKDF=t.add_type(HashTupleKDF),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Standardization utilities for Python packaging.
"""

__all__ = ["make_module", "remake_module", "remake_subpackage"]


import sys

from aiootp._typing import Typing as t

from .namespaces import FrozenNamespace


def make_module(name: str, *, mapping: dict) -> FrozenNamespace:
    """
    Turns a mapping into a module object version of a Namespace which is
    importable using normal python syntax.
    """
    module = t.ModuleType(name)
    module.__dict__.update(mapping)
    sys.modules[name] = module
    return FrozenNamespace(module.__dict__)


def remake_module(module: t.ModuleType, /) -> FrozenNamespace:
    """
    The interface for overwriting the package's modules consistently,
    applying the changes which specify the UI/UX of each.
    """
    # fmt: off
    api = module.module_api
    name = api["__name__"]  # .split(".")[-1]  # <- Uncomment for access to
                            # private variables & package debugging using
                            # from aiootp.<module> import <private_variable>
    # fmt: on
    return make_module(name=name, mapping=api)


def remake_subpackage(package: t.ModuleType, /) -> FrozenNamespace:
    """
    Applies the `remake_module` recursively to a subpackage's modules &
    the subpackage itself.
    """
    for name, subpackage in getattr(package, "subpackages", {}).items():
        package.module_api[name] = remake_subpackage(subpackage)
    for name, module in getattr(package, "modules", {}).items():
        package.module_api[name] = remake_module(module)
    return remake_module(package)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    make_module=make_module,
    remake_module=remake_module,
    remake_subpackage=remake_subpackage,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A general definition for the SyntheticIV interface.
"""

__all__ = ["SyntheticIV"]


from aiootp._typing import Typing as t
from aiootp._constants import ENCRYPTION
from aiootp._exceptions import Issue
from aiootp.commons import FrozenInstance


class SyntheticIV(FrozenInstance):
    """
    Manages the application of synthetic IVs which improve the salt
    misuse-reuse resistance of dual-output mode ciphers, since if either
    the timestamp or ephemeral SIV-key are unique, then the entire
    stream of key material will be unique. The required plaintext
    padding is handled within the `Padding` class.
    """

    __slots__ = ()

    @classmethod
    async def avalidated_transform(
        cls,
        datastream: t.AsyncDatastream,
        keystream: t.Callable[[bytes], t.Awaitable[bytes]],
        shmac: t.StreamHMACType,
    ) -> bytes:
        """
        Derives the synthetic IV from the timestamp & ephemeral SIV-key
        in the plaintext header then seeds it into the keystream to
        randomize it before encrypting / decrypting the first block of
        payload data.

        This method ciphers / deciphers the first block of plaintext /
        ciphertext depending on whether the shmac has been set to
        encryption or decryption modes.
        """
        try:
            block = await datastream.asend(None)
        except StopAsyncIteration as error:
            raise Issue.stream_is_empty() from error
        if shmac._mode == ENCRYPTION:
            return await cls._aunique_cipher(block, keystream, shmac)
        else:
            return await cls._aunique_decipher(block, keystream, shmac)

    @classmethod
    def validated_transform(
        cls,
        datastream: t.Datastream,
        keystream: t.Callable[[bytes], bytes],
        shmac: t.StreamHMACType,
    ) -> bytes:
        """
        Derives the synthetic IV from the timestamp & ephemeral SIV-key
        in the plaintext header then seeds it into the keystream to
        randomize it before encrypting / decrypting the first block of
        payload data.

        This method ciphers / deciphers the first block of plaintext /
        ciphertext depending on whether the shmac has been set to
        encryption or decryption modes.
        """
        try:
            block = datastream.send(None)
        except StopIteration as error:
            raise Issue.stream_is_empty() from error
        if shmac._mode == ENCRYPTION:
            return cls._unique_cipher(block, keystream, shmac)
        else:
            return cls._unique_decipher(block, keystream, shmac)


module_api = dict(
    SyntheticIV=t.add_type(SyntheticIV),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp._permutations.affine_permutation import AffinePermutation
from aiootp._permutations.affine_xor_chain import AffineXORChain


_key = token_bytes(2048)
_plaintexts = [*range(0, 256, 32), 255]
_test_sizes = {1, 8, 16, 32}
while len(_test_sizes) != 8:
    _test_sizes.add(choice(range(1, 33)))


class TestAffinePermutationConfig:
    kw = dict(aad=b"testing")

    async def test_size_must_be_within_bounded_limits(self) -> None:
        problem = (  # fmt: skip
            "A size out of bounds was allowed."
        )
        for bad_size in (-1, 0, 4097):
            with Ignore(ValueError, if_else=violation(problem)):
                t.AffinePermutationConfig(size=bad_size, **self.kw)

    async def test_smallest_allowed_prime(self) -> None:
        problem = (  # fmt: skip
            "A prime below 257 was allowed."
        )
        config_id = 1
        config = t.AffinePermutation._configs[config_id]
        config.is_likely_safe_multiplier(
            config.MULTIPLICATIVE_KEY, prime=257
        )
        for bad_prime in (2, 23, 59, 97, 137, 179, 251):
            with Ignore(ValueError, if_else=violation(problem)):
                config.is_likely_safe_multiplier(
                    config.MULTIPLICATIVE_KEY, prime=bad_prime
                )


class TestAffinePermutation:
    _type: type = AffinePermutation

    async def test_additive_key_size_is_enforced(self) -> None:
        problem = (  # fmt: skip
            "An invalid key size was allowed."
        )
        for size in _test_sizes:
            key = _key[: self._type.key_size(size) + 1]
            with Ignore(ValueError, if_else=violation(problem)):
                self._type(key=key, config_id=size)

    async def test_additive_key_remains_unchanged(self) -> None:
        for size in _test_sizes:
            key = _key[: self._type.key_size(size)]
            aff = self._type(key=key, config_id=size)
            assert key == _key[: aff.key_size(size)]
            assert key == aff._additive_key.to_bytes(2 * size, BIG)

    async def test_multiplicative_key_creates_correct_inverse_key(
        self,
    ) -> None:
        for size in _test_sizes:
            add_key = _key[: self._type.key_size(size)]
            aff = self._type(key=add_key, config_id=size)
            elements = [
                (aff.config.MULTIPLICATIVE_KEY * i) % aff._PRIME
                for i in _plaintexts
            ]
            assert elements != _plaintexts
            inversions = [
                (aff.config.INVERSE_KEY * i) % aff._PRIME for i in elements
            ]
            assert inversions == _plaintexts

    async def test_permutation_size_mapped_from_config_id(self) -> None:
        for size in _test_sizes:
            bitsize = 1 << (8 * size)
            key = _key[: self._type.key_size(size)]
            aff = self._type(key=key, config_id=size)
            assert bitsize == aff.config.MAX + 1
            assert size == aff.config.SIZE
            assert 2 * size == aff.key_size(size)
            assert PRIMES[(8 * size) + 1][0] == aff.config.PRIME
            assert bitsize >= aff.permute(token_bits(8 * size)) >= 0
            assert bitsize >= await aff.apermute(token_bits(8 * size)) >= 0

    async def test_permutation_is_correctly_invertible(self) -> None:
        for size in _test_sizes:
            key = _key[: self._type.key_size(size)]
            aff = self._type(key=key, config_id=size)
            permutations = [aff.permute(i) for i in _plaintexts]
            apermutations = [await aff.apermute(i) for i in _plaintexts]
            assert permutations == apermutations
            assert _plaintexts != permutations
            inversions = [aff.invert(i) for i in permutations]
            ainversions = [await aff.ainvert(i) for i in permutations]
            assert inversions == ainversions
            assert _plaintexts == inversions


class TestAffineXORChainConfig:
    kw = dict(permutation_type=AffinePermutation)

    async def test_size_must_be_within_bounded_limits(self) -> None:
        problem = (  # fmt: skip
            "A size out of bounds was allowed."
        )
        for bad_size in (-1, 0, 4097):
            with Ignore(ValueError, if_else=violation(problem)):
                t.AffineXORChainConfig(
                    config_id=32,
                    size=bad_size,
                    permutation_config_id=32,
                    **self.kw,
                )

    async def test_key_types_is_iterable_of_identifier_int_tuples(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "Invalid `key_types` argmuent value was allowed."
        )
        good_name = "in_key"
        good_int = 2
        t.AffineXORChainConfig(
            config_id=32,
            size=32,
            permutation_config_id=32,
            key_types=((good_name, good_int),),
            **self.kw,
        )
        for bad_name in ("1312non_identifier", "non-identifier"):
            with Ignore(ValueError, if_else=violation(problem)):
                t.AffineXORChainConfig(
                    config_id=32,
                    size=32,
                    permutation_config_id=32,
                    key_types=((bad_name, good_int),),
                    **self.kw,
                )
        for bad_int in (0, "zero", 2.0, "16", 33):
            with Ignore(ValueError, if_else=violation(problem)):
                t.AffineXORChainConfig(
                    config_id=32,
                    size=32,
                    permutation_config_id=32,
                    key_types=((good_name, bad_int),),
                    **self.kw,
                )

    async def test_permutation_cid_defaults_to_instance_cid(self) -> None:
        config = t.AffineXORChainConfig(
            config_id=32,
            size=32,
            **self.kw,
        )
        assert 32 == config.PERMUTATION_CONFIG_ID

    async def test_permutation_type_is_enforced(self) -> None:
        problem = (  # fmt: skip
            "An invalid permutation type was allowed."
        )

        class InvalidPermutationType:
            def permute(self, value: int) -> int:
                pass

            def invert(self, value: int) -> int:
                pass

        with Ignore(TypeError, if_else=violation(problem)):
            t.AffineXORChainConfig(
                config_id=32,
                size=32,
                permutation_type=InvalidPermutationType,
            )
        with Ignore(TypeError, if_else=violation(problem)):
            t.AffineXORChainConfig(
                config_id=32,
                size=32,
                permutation_type=None,
            )


class TestAffineXORChain:
    _type: type = AffineXORChain

    async def test_in_out_xor_key_sizes_are_enforced(self) -> None:
        problem = (  # fmt: skip
            "An invalid XOR key size was allowed."
        )
        key_reader = lambda size: size * b"A"
        bad_key_reader = lambda size: (size + 1) * b"A"
        for size in _test_sizes:
            aff = self._type.__new__(self._type)
            aff.config = self._type._configs[size]
            for tested_method in (
                aff._process_in_key,
                aff._process_out_key,
            ):
                assert aff.config.XOR_KEY_SIZE * b"A" == tested_method(
                    key_reader=key_reader, size=aff.config.XOR_KEY_SIZE
                ).to_bytes(aff.config.XOR_KEY_SIZE, BIG)

                with Ignore(ValueError, if_else=violation(problem)):
                    tested_method(key_reader=bad_key_reader, size=size)

    async def test_mid_xor_key_size_is_enforced(self) -> None:
        problem = (  # fmt: skip
            "An invalid mid XOR key size was allowed."
        )
        key_reader = lambda size: size * b"A"
        bad_key_reader = lambda size: (size + 1) * b"A"
        for size in _test_sizes:
            aff = self._type.__new__(self._type)
            aff.config = self._type._configs[size]
            tested_method = aff._process_mid_key
            IPAD, OPAD = aff.config.IPAD, aff.config.OPAD
            raw_key = int.from_bytes(aff.config.XOR_KEY_SIZE * b"A", BIG)

            assert (IPAD ^ raw_key, OPAD ^ raw_key) == tested_method(
                key_reader=key_reader, size=aff.config.XOR_KEY_SIZE
            )

            with Ignore(ValueError, if_else=violation(problem)):
                tested_method(key_reader=bad_key_reader, size=size)

    async def test_increment_update_value_must_be_within_the_domain(
        self,
    ) -> None:
        problem = (  # fmt: skip
            "An increment update value outside of the domain was allowed."
        )
        for size in _test_sizes:
            key = _key[: self._type.key_size(config_id=size)]
            aff = self._type(key=key, config_id=size)
            DOMAIN = aff.config.DOMAIN

            await aff.aupdate_increment(DOMAIN - 1)
            aff.update_increment(DOMAIN - 1)

            await aff.aupdate_increment(int(randoms.uniform(2, DOMAIN)))
            aff.update_increment(int(randoms.uniform(2, DOMAIN)))

            await aff.aupdate_increment(1)
            aff.update_increment(1)

            with Ignore(ValueError, if_else=violation(problem)):
                await aff.aupdate_increment(DOMAIN)
            with Ignore(ValueError, if_else=violation(problem)):
                aff.update_increment(DOMAIN)
            with Ignore(ValueError, if_else=violation(problem)):
                await aff.aupdate_increment(0)
            with Ignore(ValueError, if_else=violation(problem)):
                aff.update_increment(0)
            with Ignore(ValueError, if_else=violation(problem)):
                await aff.aupdate_increment(-1)
            with Ignore(ValueError, if_else=violation(problem)):
                aff.update_increment(-1)

    def recomposed_key(self, aff) -> bytes:
        size = aff.config.SIZE
        add_key_size = 2 * size
        return (
            aff._in_key.to_bytes(size, BIG)
            + (aff._in_mid_key ^ aff.config.IPAD).to_bytes(size, BIG)
            + aff._out_key.to_bytes(size, BIG)
            + aff._aff_in._additive_key.to_bytes(add_key_size, BIG)
            + aff._aff_mid._additive_key.to_bytes(add_key_size, BIG)
            + aff._aff_out._additive_key.to_bytes(add_key_size, BIG)
        )

    async def test_subpermutations_are_affine_permutation_types(
        self,
    ) -> None:
        for size in _test_sizes:
            key = _key[: self._type.key_size(config_id=size)]
            aff = self._type(key=key, config_id=size)
            for permutation in (aff._aff_in, aff._aff_mid, aff._aff_out):
                assert issubclass(permutation.__class__, AffinePermutation)

    async def test_subkeys_non_overlapping_slices_of_input_key(
        self,
    ) -> None:
        for size in _test_sizes:
            key = _key[: self._type.key_size(config_id=size)]
            aff = self._type(key=key, config_id=size)
            assert (aff._in_mid_key ^ aff.config.IPAD) == (
                aff._out_mid_key ^ aff.config.OPAD
            )
            assert key == self.recomposed_key(aff)

    async def test_invalid_key_size_is_disallowed(self) -> None:
        problem = (
            "An invalid input key size was allowed during initialization."
        )
        for size in _test_sizes:
            key = _key[: self._type.key_size(config_id=size) - 1]
            with Ignore(ValueError, if_else=violation(problem)):
                self._type(key=key, config_id=size)

            key = _key[: self._type.key_size(config_id=size) + 1]
            with Ignore(ValueError, if_else=violation(problem)):
                self._type(key=key, config_id=size)

    async def test_permutation_is_correctly_invertible(self) -> None:
        for size in _test_sizes:
            key = _key[: self._type.key_size(config_id=size)]
            aff = self._type(key=key, config_id=size)
            uncapped_permutations = [
                aff.uncapped_permute(i) for i in _plaintexts
            ]
            auncapped_permutations = [
                await aff.auncapped_permute(i) for i in _plaintexts
            ]
            permutations = [aff.permute(i) for i in _plaintexts]
            apermutations = [await aff.apermute(i) for i in _plaintexts]
            assert permutations == apermutations
            assert uncapped_permutations == auncapped_permutations
            assert _plaintexts != permutations
            assert _plaintexts != uncapped_permutations
            uncapped_inversions = [
                aff.uncapped_invert(i) for i in uncapped_permutations
            ]
            auncapped_inversions = [
                await aff.auncapped_invert(i)
                for i in auncapped_permutations
            ]
            inversions = [aff.invert(i) for i in permutations]
            ainversions = [await aff.ainvert(i) for i in apermutations]
            assert inversions == ainversions
            assert uncapped_inversions == auncapped_inversions
            assert _plaintexts == inversions
            assert _plaintexts == uncapped_inversions

    async def validate_latin_square_property(self, aff, _, DOMAIN) -> None:
        latin_square = set()
        for _ in range(DOMAIN):
            row = tuple(aff.permute(i) for i in range(DOMAIN))
            assert all((0 <= element < DOMAIN) for element in row)
            latin_square.add(row)
            await aff.astep()
        assert len(latin_square) == DOMAIN
        assert all(len(set(row)) == DOMAIN for row in latin_square)
        columns = set()
        for column_index in range(DOMAIN):
            column = tuple(row[column_index] for row in latin_square)
            columns.add(column)
        assert len(columns) == DOMAIN
        assert all(len(set(column)) == DOMAIN for column in columns)

    async def test_step_causes_full_domain_evaluation_to_fill_a_latin_square(
        self,
    ) -> None:
        SIZE = 1
        DOMAIN = 256 * SIZE
        key = _key[: self._type.key_size(config_id=SIZE)]
        aff = self._type(key=key, config_id=SIZE)
        assert aff.config.DOMAIN == DOMAIN == (aff._aff_in._MAX + 1)
        await self.validate_latin_square_property(aff, SIZE, DOMAIN)
        await aff.aupdate_increment(choice(list(range(1, DOMAIN))))
        await self.validate_latin_square_property(aff, SIZE, DOMAIN)
        """
        # Optional Extended Full-Domain Tests
        failures = []
        for i in range(256):
            try:
                aff.update_increment(i)
                await self.validate_latin_square_property(aff, SIZE, DOMAIN)
            except AssertionError as error:
                failures.append(i)
        assert not any(failures), failures
        """


class TestFastAffineXORChain(TestAffineXORChain):
    _type: type = FastAffineXORChain

    def recomposed_key(self, aff) -> bytes:
        size = aff.config.SIZE
        add_key_size = 2 * size
        assert aff._aff_in is aff._aff_mid
        assert aff._aff_mid is aff._aff_out
        return (
            aff._in_key.to_bytes(size, BIG)
            + (aff._in_mid_key ^ aff.config.IPAD).to_bytes(size, BIG)
            + aff._out_key.to_bytes(size, BIG)
            + aff._aff_in._additive_key.to_bytes(add_key_size, BIG)
        )


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Types for creating & reading unique unblinded IDs, or IDs blinded
with bijective, keyed permutations.
"""

__all__ = ["GUID", "SequenceID"]


from .sequence_id_config import *
from .sequence_id import *
from .raw_guid_config import *
from .raw_guid import *
from .guid_config import *
from .guid import *


modules = dict(
    guid=guid,
    guid_config=guid_config,
    raw_guid=raw_guid,
    raw_guid_config=raw_guid_config,
    sequence_id=sequence_id,
    sequence_id_config=sequence_id_config,
)


module_api = dict(
    GUID=GUID,
    SequenceID=SequenceID,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A multi-processing interface.
"""

__all__ = ["Processes", "get_process_id"]


import multiprocessing
from os import getpid as get_process_id
from concurrent.futures import process
from concurrent.futures import ProcessPoolExecutor

from aiootp._typing import Typing as t

from .concurrency_interface import ConcurrencyInterface


class Processes(ConcurrencyInterface):
    """
    Simplifies spawning & returning the values procuded by `Process` &
    `ProcessPoolExecutor` objects with an interface over these types
    from the `multiprocessing` & `concurrent.futures` packages.
    """

    __slots__ = ()

    _Manager: type = multiprocessing.Manager

    _default_probe_delay: t.PositiveRealNumber = 0.005
    _pool: t.PoolExecutorType = ProcessPoolExecutor()
    _type: type = multiprocessing.Process

    BrokenPool: type = process.BrokenProcessPool


module_api = dict(
    Processes=t.add_type(Processes),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    get_process_id=get_process_id,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from collections import deque

from conftest import *

from aiootp._constants.misc import HASHER_TYPES


class BlankType:
    pass


class ProtocolSubTypeTests:
    async def test_issubclass_at_runtime(self) -> None:
        for type_tested in self.types_tested:
            assert issubclass(type_tested, self.protocol)
            assert not issubclass(BlankType, self.protocol)

    async def test_isinstance_at_runtime(self) -> None:
        for instance_tested in self.instances_tested:
            assert isinstance(instance_tested, self.protocol)
            assert not isinstance(BlankType(), self.protocol)


class TestPaddingType(ProtocolSubTypeTests):
    protocol = t.PaddingType
    types_tested = [t.Padding]
    (config, *_) = randoms.choice(all_ciphers)
    instances_tested = [t.Padding(config)]


class TestStreamHMACTypes(ProtocolSubTypeTests):
    protocol = t.StreamHMACType
    types_tested = []
    instances_tested = []
    for _, cipher, *_ in all_ciphers:
        types_tested.append(cipher._StreamHMAC)
        instances_tested.append(cipher.stream_encrypt().shmac)

    async def test_hasattr_on_properties(self) -> None:
        for instance_tested in self.instances_tested:
            hasattr(instance_tested, "result")


class TestSyntheticIVTypes(ProtocolSubTypeTests):
    protocol = t.SyntheticIVType
    types_tested = []
    instances_tested = []
    for _, cipher, *_ in dual_output_ciphers:
        typ = cipher._Junction._SyntheticIV
        types_tested.append(typ)
        instances_tested.append(typ())


class TestAsyncCipherStreamingTypes(ProtocolSubTypeTests):
    protocol = t.AsyncCipherStreamingType
    types_tested = [t.AsyncCipherStream, t.AsyncDecipherStream]
    (config, cipher, salt, _) = randoms.choice(all_ciphers)
    # fmt: off
    instances_tested = [
        run(t.AsyncCipherStream(cipher)),
        run(t.AsyncDecipherStream(
            cipher, salt=salt, iv=csprng(config.IV_BYTES)
        )),
    ]
    # fmt: on


class TestCipherStreamingTypes(ProtocolSubTypeTests):
    protocol = t.CipherStreamingType
    types_tested = [t.CipherStream, t.DecipherStream]
    (config, cipher, salt, _) = randoms.choice(all_ciphers)
    instances_tested = [
        t.CipherStream(cipher),
        t.DecipherStream(cipher, salt=salt, iv=csprng(config.IV_BYTES)),
    ]


class TestCipherInterfaceTypes(ProtocolSubTypeTests):
    protocol = t.CipherInterfaceType
    types_tested = [cipher.__class__ for (_, cipher, *_) in all_ciphers]
    instances_tested = [type_tested(key) for type_tested in types_tested]


class TestConfigTypes(ProtocolSubTypeTests):
    protocol = t.ConfigType
    types_tested = [config.__class__ for (config, *_) in all_ciphers]
    instances_tested = [config for (config, *_) in all_ciphers]


class TestSupportsPopleftTypes(ProtocolSubTypeTests):
    protocol = t.SupportsPopleft
    types_tested = [deque]
    instances_tested = [type_tested([0, 1]) for type_tested in types_tested]


class TestSupportsAppendPopTypes(ProtocolSubTypeTests):
    protocol = t.SupportsAppendPop
    types_tested = [deque, list]
    instances_tested = [type_tested([0, 1]) for type_tested in types_tested]


class TestSupportsAppendPopleftTypes(ProtocolSubTypeTests):
    protocol = t.SupportsAppendPopleft
    types_tested = [deque]
    instances_tested = [type_tested([0, 1]) for type_tested in types_tested]


class TestHasherTypes(ProtocolSubTypeTests):
    protocol = t.HasherType
    types_tested = [
        config.factory().__class__ for config in HASHER_TYPES.values()
    ]
    instances_tested = [
        config.factory() for config in HASHER_TYPES.values()
    ]


class TestXOFTypes(ProtocolSubTypeTests):
    protocol = t.XOFType
    types_tested = [
        config.factory().__class__
        for config in HASHER_TYPES.values()
        if config.factory().digest_size == 0
    ]
    instances_tested = [
        config.factory()
        for config in HASHER_TYPES.values()
        if config.factory().digest_size == 0
    ]


class TestPermutationTypes(ProtocolSubTypeTests):
    protocol = t.PermutationType
    types_tested = [
        t.AffinePermutation,
        t.AffineXORChain,
        t.FastAffineXORChain,
    ]
    instances_tested = [
        type_tested(key=csprng(type_tested.key_size(16)), config_id=16)
        for type_tested in types_tested
    ]


class TestPoolExecutorTypes(ProtocolSubTypeTests):
    protocol = t.PoolExecutorType
    types_tested = [t.Processes._pool.__class__, t.Threads._pool.__class__]
    instances_tested = [t.Processes._pool, t.Threads._pool]


class TestAsyncDatabaseType(ProtocolSubTypeTests):
    protocol = t.AsyncDatabaseType
    types_tested = [AsyncDatabase]
    instances_tested = [
        run(type_tested(key)) for type_tested in types_tested
    ]


class TestDatabaseType(ProtocolSubTypeTests):
    protocol = t.DatabaseType
    types_tested = [Database]
    instances_tested = [type_tested(key) for type_tested in types_tested]


class TestDomainKDFTypes(ProtocolSubTypeTests):
    protocol = t.DomainKDFType
    types_tested = [DomainKDF, t.DBKDF]
    instances_tested = [
        type_tested(b"test-domain", key=key) for type_tested in types_tested
    ]


class TestPublicKeyTypes(ProtocolSubTypeTests):
    protocol = t.PublicKeyType
    instances_tested = [
        typ().generate().public_key for typ in (X25519, Ed25519)
    ]
    types_tested = [
        instance_tested.__class__ for instance_tested in instances_tested
    ]


class TestSecretKeyTypes(ProtocolSubTypeTests):
    protocol = t.SecretKeyType
    instances_tested = [
        typ().generate().secret_key for typ in (X25519, Ed25519)
    ]
    types_tested = [
        instance_tested.__class__ for instance_tested in instances_tested
    ]


class TestAsymmetricKeyTypes(ProtocolSubTypeTests):
    protocol = t.AsymmetricKeyType
    instances_tested = [
        X25519(),
        Ed25519(),
        X25519().generate(),
        Ed25519().generate(),
    ]
    types_tested = [
        instance_tested.__class__ for instance_tested in instances_tested
    ]


class TestSignerTypes(ProtocolSubTypeTests):
    protocol = t.SignerType
    instances_tested = [Ed25519(), Ed25519().generate()]
    types_tested = [
        instance_tested.__class__ for instance_tested in instances_tested
    ]


class TestKeyExchangeTypes(ProtocolSubTypeTests):
    protocol = t.KeyExchangeType
    instances_tested = [X25519(), X25519().generate()]
    types_tested = [
        instance_tested.__class__ for instance_tested in instances_tested
    ]


class TestKeyExchangeProtocolTypes(ProtocolSubTypeTests):
    protocol = t.KeyExchangeProtocolType
    key = X25519()
    instances_tested = [
        X25519.dh2_client(),
        key.dh2_server(),
        key.dh3_client(),
        key.dh3_server(),
    ]
    types_tested = [
        instance_tested.__class__ for instance_tested in instances_tested
    ]


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Configuration type for `SequenceID`.
"""

__all__ = ["SequenceIDConfig"]


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue
from aiootp.commons import Config


class SequenceIDConfig(Config):
    """
    Configuration type for `SequenceID`.
    """

    __slots__ = ("SIZE", "KEY_SIZE", "PERMUTATION_CONFIG_ID", "Permutation")

    slots_types: t.Mapping[str, type] = dict(
        SIZE=int,
        KEY_SIZE=int,
        PERMUTATION_CONFIG_ID=t.Hashable,
        Permutation=t.PermutationType,
    )

    def _process_size(self, size: int) -> int:
        """
        Returns `size` only if it passes value-based validity checks,
        otherwise raises `ValueError`.
        """
        if 4096 >= size > 0:
            return size
        else:
            raise Issue.value_must("size", "be > 0 and <= 4096")

    def _process_permutation_config_id(
        self, permutation_config_id: t.Optional[t.Hashable]
    ) -> t.Hashable:
        """
        Allows the configuration ID given to the initializer of the
        permutation type to be specified. If it's `None` then it's
        assumed the correct configuration ID is the same used by this
        instance. Returns the configuration ID that's selected.
        """
        if permutation_config_id is None:
            return self.CONFIG_ID
        else:
            return permutation_config_id

    def __init__(
        self,
        *,
        config_id: t.Hashable,
        size: int,
        permutation_type: t.PermutationType,
        permutation_config_id: t.Optional[t.Hashable] = None,
    ) -> None:
        self.CONFIG_ID = config_id
        self.SIZE = self._process_size(size)
        self.PERMUTATION_CONFIG_ID = self._process_permutation_config_id(
            permutation_config_id
        )
        self.Permutation = permutation_type
        self.KEY_SIZE = self.Permutation.key_size(config_id)


module_api = dict(
    SequenceIDConfig=t.add_type(SequenceIDConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `asynchs` subpackage.
"""

__all__ = ["AsyncOrSyncIterable", "ClockType", "Future", "PoolExecutorType"]


from concurrent.futures._base import Future

from .interface import Typing as t


class _AsyncOrSyncIterableMeta(type):
    """
    Allows bracketed choices of types to be given to the `Iterable` &
    `AsyncIterable` type hinters for the `AsyncOrSyncIterable` subclass.
    """

    def __getitem__(cls, obj: t.Any) -> t.Union:
        return t.Union[t.Iterable[obj], t.AsyncIterable[obj]]


class AsyncOrSyncIterable(metaclass=_AsyncOrSyncIterableMeta):
    """
    Allows bracketed choices of types to be given to the `Iterable` &
    `AsyncIterable` type hinters.
    """


@t.runtime_checkable
class PoolExecutorType(t.Protocol):
    def map(
        self,
        fn: t.Callable[..., t.Any],
        *iterables: t.Any,
        timeout: t.Optional[t.PositiveRealNumber],
        chunksize: int,
    ) -> t.Iterator[t.Any]:
        pass  # pragma: no cover

    def shutdown(self, wait: bool) -> None:
        pass  # pragma: no cover

    def submit(
        self,
        fn: t.Callable[..., t.Any],
        /,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> Future:
        pass  # pragma: no cover


@t.runtime_checkable
class ClockType(t.Protocol):
    async def atime(self) -> int:
        pass  # pragma: no cover

    def time(self) -> int:
        pass  # pragma: no cover

    async def amake_timestamp(self, *, size: int, byte_order: str) -> bytes:
        pass  # pragma: no cover

    def make_timestamp(self, *, size: int, byte_order: str) -> bytes:
        pass  # pragma: no cover

    async def atest_timestamp(
        self, timestamp: bytes, ttl: t.Optional[int], *, byte_order: str
    ) -> None:
        pass  # pragma: no cover

    def test_timestamp(
        self, timestamp: bytes, ttl: t.Optional[int], *, byte_order: str
    ) -> None:
        pass  # pragma: no cover


module_api = dict(
    AsyncOrSyncIterable=t.add_type(AsyncOrSyncIterable),
    ClockType=t.add_type(ClockType),
    Future=t.add_type(Future),
    PoolExecutorType=t.add_type(PoolExecutorType),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class ExampleType:
    pass


class FalseConfig:
    def set_config_id(self, config_id: t.Hashable) -> None:
        self.CONFIG_ID = config_id


class TestConfig:
    types_tested = (int, str, bytes, list, dict, ExampleType)
    uncheckable_types = (t.List[int], t.PositiveRealNumber)

    def test_slots_types_enforces_types(self) -> None:
        problem = (  # fmt: skip
            "A mismatch between an attribute type & a `slot_types` "
            "delcaration was allowed."
        )
        for declaration in self.types_tested:

            class ExampleConfig(Config):
                __slots__ = ("VAR",)
                slots_types = dict(VAR=declaration)

                def __init__(self, *, var: declaration) -> None:
                    self.VAR = var

            for attr_type in set(self.types_tested).difference(
                [declaration]
            ):
                with Ignore(TypeError, if_else=violation(problem)):
                    ExampleConfig(var=attr_type)

    def test_runtime_uncheckable_types_are_detected(self) -> None:
        problem = (  # fmt: skip
            "A runtime-uncheckable type declaration didn't proc an error."
        )
        for declaration in self.uncheckable_types:
            with Ignore(
                TypeUncheckableAtRuntime, if_else=violation(problem)
            ):

                class ExampleConfig(Config):
                    __slots__ = ("VAR",)
                    slots_types = dict(VAR=declaration)

                    def __init__(self, *, var: declaration) -> None:
                        self.VAR = var


class TestConfigMap:
    def test_mapping_registers_config_by_id(
        self, mapping: ConfigMap
    ) -> None:
        for config_id in ("one", b"one", 1):
            config = ExampleConfig(number=420, string="word")

            mapping[config_id] = config
            assert config_id == config.CONFIG_ID
            assert config is mapping[config_id]

    def test_config_id_cannot_change(
        self, config: t.ConfigType, mapping: ConfigMap
    ) -> None:
        problem = (  # fmt: skip
            "config_id changed without error."
        )
        config_id = 1
        mapping[config_id] = config
        with Ignore(ValueError, if_else=violation(problem)):
            config.CONFIG_ID = 2

        assert config.CONFIG_ID == 1

    def test_config_must_be_config_subclass(
        self, mapping: ConfigMap
    ) -> None:
        problem = (  # fmt: skip
            "non-`Config` subclass was allowed to be registered."
        )
        config_id = 1
        with Ignore(TypeError, if_else=violation(problem)):
            ConfigMap(config_type=FalseConfig)
        with Ignore(TypeError, if_else=violation(problem)):
            mapping[config_id] = FalseConfig()

    def test_cannot_be_reassigned(
        self, config: t.ConfigType, mapping: ConfigMap
    ) -> None:
        problem = (  # fmt: skip
            "a config was allowed to be reassigned to the map."
        )
        config_id = 1
        mapping[config_id] = config
        with Ignore(PermissionError, if_else=violation(problem)):
            mapping[config_id] = ExampleConfig(number=111, string="abd")

    def test_reassigning_doesnt_throw_when_same_object(
        self, config: t.ConfigType, mapping: ConfigMap
    ) -> None:
        config_id = 1
        mapping[config_id] = config
        mapping[config_id] = config


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Manager interfaces for cipher KDFs.
"""

__all__ = ["CipherKDFs", "KeyedCipherKDF", "ShakeConfig"]


from aiootp._typing import Typing as t
from aiootp._constants.misc import MIN_KEY_BYTES
from aiootp._exceptions import KeyAADIssue, Issue
from aiootp.commons import Config, FrozenInstance
from aiootp.generics import Domains
from aiootp.generics import encode_key, fullblock_ljust


class ShakeConfig(Config):
    """
    Allows for a domain-separated initialization of a cipher's KDF.
    """

    __slots__ = (
        "name",
        "pad",
        "offset_amount",
        "offset",
        "blocksize",
        "double_blocksize",
        "hasher_type",
        "hasher",
        "factory",
        "key_slice",
    )

    slots_types: t.Mapping[str, type] = dict(
        name=str,
        pad=bytes,
        offset_amount=int,
        offset=bytes,
        blocksize=int,
        double_blocksize=int,
        hasher_type=type,
        hasher=t.Callable,
        factory=t.Callable,
        key_slice=slice,
    )

    def __init__(
        self,
        name: str,
        pad: bytes,
        offset_amount: int,
        hasher: t.Callable,
        key_slice: t.Optional[slice],
    ) -> None:
        self.name = name
        self.pad = pad
        self.offset_amount = offset_amount
        self.offset = offset_amount * pad
        self.hasher = hasher
        self.hasher_type = type(hasher())
        self.blocksize = hasher().block_size
        self.double_blocksize = 2 * self.blocksize
        self.key_slice = key_slice

    def prepare_factory(
        self, config_id: bytes, packed_metadata: bytes
    ) -> t.Self:
        """
        Domain separates each KDF with a unique salt based on the
        relative name of the KDF & the configuration metadata of the
        cipher.
        """
        self.factory = self.hasher(
            Domains.encode_constant(
                f"{self.name}_salt",
                domain=config_id,
                aad=packed_metadata,
                size=self.double_blocksize,
            )
        ).copy
        return self


class KeyedCipherKDF(FrozenInstance):
    """
    Allows the amortization of an initialized cipher's KDF, allowing the
    efficient use of the cipher with arbitrarily large sized keys.
    """

    __slots__ = ("config", "kdf")

    def __init__(self, key: bytes, config: ShakeConfig) -> None:
        self.config = config
        self.kdf = config.factory()
        self.kdf.update(encode_key(key, config.blocksize, pad=config.pad))

    def new_session_copy(self, summary: bytes) -> t.XOFType:
        """
        Returns a copy of the keyed KDF after being initialized with the
        session's `salt`, `aad`, & `iv` canonically encoded into a
        `summary`.
        """
        config = self.config
        session_kdf = self.kdf.copy()
        session_kdf.update(
            fullblock_ljust(summary, config.blocksize, pad=config.pad)
            + config.offset
        )
        return session_kdf


class CipherKDFs(FrozenInstance):
    """
    A private type which is responsible for initializing the keystream
    key-derivation objects for chained sponge ciphers.
    """

    __slots__ = ("config",)

    def test_key_validity(self, key: bytes) -> None:
        """
        Throws exception if `key` violates type & min size requirements.
        """
        if key.__class__ is not bytes:
            raise Issue.value_must_be_type("key", bytes)
        elif len(key) < MIN_KEY_BYTES:
            raise KeyAADIssue.invalid_key_size(len(key), MIN_KEY_BYTES)

    def key_base_kdf(self, name: str, *, key: bytes) -> KeyedCipherKDF:
        """
        Keys & returns an initialized KDF.
        """
        return KeyedCipherKDF(key, self.config.KDF_CONFIGS[name])


module_api = dict(
    CipherKDFs=t.add_type(CipherKDFs),
    KeyedCipherKDF=t.add_type(KeyedCipherKDF),
    ShakeConfig=t.add_type(ShakeConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)
